import os
import sys

if "LOGOS_RECOGNITION" in os.environ:
    PATH_GLOB = os.environ["LOGOS_RECOGNITION"]
else:
    PATH_GLOB = os.path.expanduser(os.path.join(
        os.environ["HOME"], ".hkt", "logodetect"))

# We append the base path for constants.py to the Python path dynamically,
# so we can import it project-wide.
sys.path.append(os.path.abspath(PATH_GLOB))



"""Augmentations for outdoor environments.
"""

import itertools
import numpy as np
from PIL import Image
from imgaug import augmenters

from constants import AUGMENTER_PARAMS


def get_augmentations(image: Image.Image, params: dict = AUGMENTER_PARAMS) -> list:
    """Get a list of augmented images for an image,
    according to the specified augmentation

    :param image: a PIL.Image instance
    :param params: dictionary of augmenter params.
    :return: list of augmented PIL.Image images
    """
    param_product = itertools.product(*params.values())
    return [augment_image(image, *param) for param in param_product]


def augment_image(
    image: Image.Image, mu: float, sigma: float, scale: float, shear: int, rotate: int
) -> Image.Image:
    """Augment an image using the 'imgaug' library.

    :param image: a PIL.image instance
    :param mu: Gaussian blur mu
    :param sigma: Gaussian blur sigma
    :param scale: Additive Gaussian noise
    :param shear: Affine shear
    :param rotate: Affine rotation
    :return: Augmented PIL.Image
    """
    augmenter = augmenters.Sequential(
        [
            augmenters.Multiply(mul=mu),
            augmenters.GaussianBlur(sigma=sigma),
            augmenters.AdditiveGaussianNoise(scale=scale),
            augmenters.Affine(rotate=rotate, shear=shear),
        ]
    )
    image_arr = augmenter.augment_image(np.array(image))
    return Image.fromarray(image_arr)



import sys

import numpy as np
import pandas as pd
from PIL import Image
import torch
from torch.nn import functional
import torchvision

from logodetect import classifiers
from logodetect.augmenters import get_augmentations
from logodetect.utils import (
    open_and_resize,
    clean_name,
)
from constants import (
    CLASSIFIER_ALG,
    PATH_EXEMPLARS_EMBEDDINGS,
    EMBEDDER_ALG,
    EMBEDDER_WEIGHTS,
    EMBEDDER_DEVICE,
    BRAND_LOGOS,
    IMAGE_RESIZE,
    LOAD_EMBEDDINGS,
    EMBEDDING_SIZE,
    DISTANCE,
    MAX_DISTANCE,
)


class Classifier:
    """KNN Classifier
    """

    def __init__(self, exemplar_paths: str) -> None:

        self.transform = self._compute_transform()
        self.embedder = classifiers.get(EMBEDDER_ALG)(EMBEDDER_DEVICE, EMBEDDER_WEIGHTS)
        self._load_exemplars(exemplar_paths)
        self.classifier = self._set_classifier()

    @staticmethod
    def _compute_transform() -> torchvision.transforms.Compose:
        """Compute image transform for inference.

        :return: torchvision transform
        """
        return torchvision.transforms.Compose(
            [
                torchvision.transforms.Resize(IMAGE_RESIZE),
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                ),
            ]
        )

    def _load_exemplars(self, exemplars_paths: str) -> None:
        """Load exemplars from file

        :param exemplars_paths: path to exemplars folder
        :return: None
        """
        if LOAD_EMBEDDINGS:
            # Load embeddings and names of exemplars:
            exemplars = pd.read_pickle(PATH_EXEMPLARS_EMBEDDINGS)
            mask = exemplars["brand"].isin(BRAND_LOGOS)
            self.exemplars_vectors = list(exemplars["img_vec"][mask])
            self.exemplars_brands = list(exemplars["brand"][mask])
        else:
            self.exemplars_vectors = []
            self.exemplars_brands = []
            for path in exemplars_paths:
                brand = clean_name(path)
                image = open_and_resize(path, IMAGE_RESIZE)

                # Store clean image:
                embedding = self.embed_image(image)
                self.exemplars_vectors.append(embedding)
                self.exemplars_brands.append(brand)

                # Store augmented image:
                for aug_image in get_augmentations(image):
                    embedding = self.embed_image(aug_image)
                    self.exemplars_vectors.append(embedding)
                    self.exemplars_brands.append(brand)

    def _set_classifier(self):
        """Determine parameters for scikit-learn KNeighborsClassifier, fit
        data to it and return it.

        :return: trained classifier model
        """
        if CLASSIFIER_ALG is not "knn":
            raise ValueError(
                f"A classifiers.knn.Classifier can only be run with CLASSIFIER_ALG='knn', got {CLASSIFIER_ALG}."
            )
        if DISTANCE.lower() == "minkowski_1":
            model = classifiers.get(CLASSIFIER_ALG)(
                n_neighbors=1, metric="minkowski", p="1"
            )
        elif DISTANCE.lower() == "minkowski_2":
            model = classifiers.get(CLASSIFIER_ALG)(
                n_neighbors=1, metric="minkowski", p="2"
            )
        elif DISTANCE.lower() == "cosine":
            model = classifiers.get(CLASSIFIER_ALG)(n_neighbors=1, metric="cosine")
        else:
            print("{} is not a valid distance.".format(DISTANCE))
            sys.exit()
        # TODO: code smell here, setting a classifier does not suggest training (or returning anything)
        return model.fit(self.exemplars_vectors, self.exemplars_brands)

    def predict(self, detections: dict, image: np.ndarray) -> dict:
        """Classify an image given precomputed detection objects

        :param detections: detections dictionary with overlay boxes
        :param image: input image
        :return: augmented detections
        """
        if len(detections["boxes"]) != 0:
            image = Image.fromarray(image)
            detections = self.load_detections(image, detections)
            detections = self.classify_embeddings(detections)
        return detections

    def load_detections(self, image: Image.Image, detections: dict) -> dict:
        """Load all detections specific to the given input image

        :param image: input image
        :param detections: precomputed detections
        :return: detections
        """
        detections_mat = np.zeros((len(detections["boxes"]), EMBEDDING_SIZE))
        for idx, box in enumerate(detections["boxes"]):
            crop_image = image.crop(box).resize(IMAGE_RESIZE)
            embedding = self.embed_image(crop_image)
            detections_mat[idx, :] = embedding
        detections["embeddings"] = detections_mat
        return detections

    def embed_image(self, image):
        """Embeds an image using the embedding algorithm

        :param image: input image
        :return: normalized embedding
        """
        image = (
            self.transform(image).unsqueeze(0).to(EMBEDDER_DEVICE, dtype=torch.float)
        )
        embedding = self.embedder(image)
        normalized_embedding = functional.normalize(embedding, p=2, dim=1)
        return normalized_embedding.squeeze().detach().cpu().numpy()

    def classify_embeddings(self, detections: dict) -> dict:
        """Use the classifier algorithm to predict classes of
        the given detections.

        :param detections:
        :return:
        """
        dists, classes = self.classifier.kneighbors(detections["embeddings"])
        brands = [self.exemplars_brands[idx] for idx in classes.flatten()]
        detections["brands"] = np.array(brands)
        selections = (dists < MAX_DISTANCE).flatten()
        return self._select_detections(detections, selections)

    @staticmethod
    def _select_detections(detections, selections):
        """Specify detections for the given selections.

        :param detections:
        :param selections:
        :return:
        """
        detections["boxes"] = detections["boxes"][selections]
        detections["labels"] = detections["labels"][selections]
        detections["scores"] = detections["scores"][selections]
        detections["brands"] = detections["brands"][selections]
        return detections



"""Detector module."""

# Standard library:
import os

# Pip packages:
import numpy as np
import torch

# Current library:
from logodetect import detectors
from logodetect.utils import image_to_gpu_tensor
from constants import (
    DETECTOR_DEVICE,
    DETECTOR_ALG,
    DETECTOR_WEIGHTS,
    MIN_CONFIDENCE,
)


class Detector:
    "Add documentation."

    def __init__(self):
        model = detectors.get(DETECTOR_ALG)
        self.model = model(DETECTOR_DEVICE, DETECTOR_WEIGHTS)

    @torch.no_grad()
    def predict(self, image):
        "Add documentation."
        image = image_to_gpu_tensor(image, DETECTOR_DEVICE)
        # Always returns list with one element:
        detections = self.model(image)[0]
        return self._process_detections(detections)

    def _process_detections(self, detections):
        "Add documentation."
        # Move to cpu:
        detections = self._detections_to_cpu(detections)
        # keep the ones above some (potential) threshold
        selections = np.array(detections["scores"] > MIN_CONFIDENCE)
        return self._select_detections(detections, selections)

    def _detections_to_cpu(self, detections):
        "Moves all the fields of 'detections' to the CPU."
        detections["boxes"] = detections["boxes"].cpu().numpy()
        detections["labels"] = detections["labels"].cpu().numpy()
        detections["scores"] = detections["scores"].cpu().numpy()
        return detections

    def _select_detections(self, detections, selections):
        "Add documentation."
        detections["boxes"] = detections["boxes"][selections]
        detections["labels"] = detections["labels"][selections]
        detections["scores"] = detections["scores"][selections]
        detections["brands"] = []
        return detections



"""Command line interface
"""
import os
import click
from functools import partial
import requests
import zipfile

# Note that this little dance is needed to dynamically load "constants.py" at run-time
# from without the repo (for users that don't clone it, but solely use pip).
try:
    import constants
except:
    import backup_constants as constants


out = partial(click.secho, bold=True, err=True)

if "LOGOS_RECOGNITION" in os.environ:
    BASE_PATH = os.environ["LOGOS_RECOGNITION"]
else:
    BASE_PATH = os.path.expanduser(os.path.join('~', '.hkt', 'logodetect'))
DATA_PATH = os.path.join(BASE_PATH, "data")
MODEL_PATH = os.path.join(BASE_PATH, "models")
BASE_URL = "https://hkt-logodetect.s3.eu-central-1.amazonaws.com"


def common_options(function):
    function = click.option(
        "-e",
        "--exemplars",
        default=constants.PATH_EXEMPLARS,
        required=False,
        help="path to your exemplars folder",
    )(function)
    function = click.option(
        "-o",
        "--output_appendix",
        default="_output",
        required=False,
        help="string appended to your resulting file",
    )(function)
    return function


@click.command()
@click.option(
    "-i",
    "--image_filename",
    default=constants.IMAGE_FILENAME,
    required=False,
    help="path to your input image",
)
@common_options
def image(image_filename, exemplars, output_appendix):
    from logodetect.recognizer import Recognizer

    recognizer = Recognizer(exemplars)
    recognizer.predict_image(
        image_filename=image_filename, output_appendix=output_appendix
    )
    out("All done! ✨ 🍰 ✨")


@click.command()
@click.option(
    "-v",
    "--video_filename",
    default=constants.VIDEO_FILENAME,
    required=False,
    help="path to your input video",
)
@common_options
def video(video_filename, exemplars, output_appendix):
    from logodetect.recognizer import Recognizer

    recognizer = Recognizer(exemplars)
    recognizer.predict(video_filename=video_filename,
                       output_appendix=output_appendix)
    out("All done! ✨ 🍰 ✨")


@click.command()
def init():
    os.makedirs(MODEL_PATH, exist_ok=True)
    print(
        ">>> Note that downloading all model and data files might take a few minutes!"
    )

    download("detector.pth", "model")
    download("embedder.pth", "model")
    download("classifier_resnet18.pth", "model")
    download("data.zip", "data")


def download(file_name: str, data_type: str):
    print(f">>> Downloading file {file_name} now, please wait...")

    if data_type == "model":
        local_path = os.path.join(MODEL_PATH, file_name)
    elif data_type == "data" or data_type == "base":
        local_path = os.path.join(BASE_PATH, file_name)
    else:
        raise ValueError(
            f"Data type {data_type} not allowed for downloads, pick from 'model', 'data' or "
            f"'base'."
        )

    # only download data if it does not already exist
    if not (
        (data_type == "data" and os.path.exists(DATA_PATH)) or
        (data_type != "data" and os.path.exists(local_path))
    ):
        url = os.path.join(BASE_URL, file_name)
        get_request = requests.get(url)
        open(local_path, "wb").write(get_request.content)
        if data_type == "data":
            with zipfile.ZipFile(local_path, "r") as zip_ref:
                zip_ref.extractall(DATA_PATH)
            os.remove(local_path)
        print(">>> Download complete!")
    else:
        print(">>> Skipping download, file already exists.")


@click.group()
def cli():
    pass


cli.add_command(init)
cli.add_command(image)
cli.add_command(video)



"""Collection of models for detection."""
from .faster_rcnn import *
import torch
from torchvision.models.detection import fasterrcnn_resnet50_fpn
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor

ALL_ARCHITECTURES = [
    "fasterrcnn_resnet50",
    "binary_fasterrcnn_resnet50",
]


def get(function_name: str):
    """Get a model architecture by function name.

    :param function_name: function name to call
    :return: the function created from the name.
    """
    if function_name not in ALL_ARCHITECTURES:
        raise ValueError(
            f"{function_name} is not a valid model architecture, choose from {ALL_ARCHITECTURES}."
        )
    return globals()[f"_{function_name}"]


def _fasterrcnn_resnet50(device: str, model_weights: str = None):
    """Creates a FasterRCNN with a ResNet50 backbone.
    """
    model = fasterrcnn_resnet50_fpn()

    if model_weights:
        checkpoint = torch.load(model_weights, map_location=device)
        model.load_state_dict(checkpoint["state_dict"])

    return model.eval().to(device)


def _binary_fasterrcnn_resnet50(device: str, model_weights: str):
    """Loads a pre-trained FasterRCNN with a ResNet50 backbone and 2-class output.
    """
    model = fasterrcnn_resnet50_fpn()

    # Get the number of input features for the top layer:
    in_features = model.roi_heads.box_predictor.cls_score.in_features
    # Replace the head with a new one for just 2 classes: background and logo
    model.roi_heads.box_predictor = FastRCNNPredictor(in_features, 2)

    # Define the computing device explicitly:
    checkpoint = torch.load(model_weights, map_location=device)
    # checkpoint = torch.load(model_weights)
    model.load_state_dict(checkpoint["state_dict"])

    return model.eval().to(device)



"""Classifier module."""

# Pip packages:
import numpy as np
from PIL import Image
import torch

# Current library:
from logodetect import classifiers
from logodetect.augmenters import get_augmentations
from logodetect.utils import clean_name, open_and_resize, image_to_gpu_tensor
from constants import (
    CLASSIFIER_ALG,
    CLASSIFIER_DEVICE,
    CLASSIFIER_WEIGHTS,
    IMAGE_RESIZE,
    MIN_CONFIDENCE,
)


class Classifier:
    """Siamese Network classifier.
    """

    def __init__(self, exemplar_paths):
        "Add documentation."
        # Define class variables:
        self.exemplars_imgs = None
        self.exemplars_brands = None

        # Set the network to classify the detections:
        self.load_exemplars(exemplar_paths)
        self.classifier = classifiers.get(CLASSIFIER_ALG)(
            CLASSIFIER_DEVICE, CLASSIFIER_WEIGHTS
        )

    def load_exemplars(self, exemplars_paths):
        "Add documentation."
        self.exemplars_imgs = []
        self.exemplars_brands = []
        for path in exemplars_paths:
            brand = clean_name(path)
            image = open_and_resize(path, IMAGE_RESIZE)

            # Store clean image:
            image_gpu = image_to_gpu_tensor(image, CLASSIFIER_DEVICE)
            self.exemplars_imgs.append(image_gpu)
            self.exemplars_brands.append(brand)

            # Store augmented image:
            for aug_image in get_augmentations(image):
                aug_image_gpu = image_to_gpu_tensor(aug_image, CLASSIFIER_DEVICE)
                self.exemplars_imgs.append((aug_image_gpu))
                self.exemplars_brands.append(brand)

    def predict(self, detections, image):
        "Add documentation."
        if len(detections["boxes"]) != 0:
            image = Image.fromarray(image)
            # Compare each detection to each exemplar in one forward pass:
            comb_images = self._create_comb_images(image, detections)
            comb_scores = self.classifier(comb_images).detach().cpu().numpy()
            detections = self._process_scores(comb_scores, detections)
        return detections

    def _create_comb_images(self, image, detections):
        """
        comb_images: Is a batch storing all combinations
                     of detections and exemplars and will have a size of
                     (n_detections * n_exemplars, 6, H, W)
        """
        comb_images = []
        for box in detections["boxes"]:
            crop = image.crop(box).resize(IMAGE_RESIZE)
            detection = image_to_gpu_tensor(crop, CLASSIFIER_DEVICE)

            for exemplar in self.exemplars_imgs:
                comb_images.append(torch.cat((detection, exemplar), 1))

        return torch.cat(comb_images).to(CLASSIFIER_DEVICE)

    def _process_scores(self, comb_scores, detections):
        "Add documentation."
        n_detections = len(detections["boxes"])
        n_logos = len(self.exemplars_imgs)
        selections = np.zeros(n_detections).astype(np.bool)

        for n_det in range(n_detections):
            a = n_det * n_logos
            b = a + n_logos
            scores = comb_scores[a:b]

            if (scores >= MIN_CONFIDENCE).any():
                detections["labels"][n_det] = np.argmax(scores)
                detections["scores"][n_det] = np.max(scores)
                selections[n_det] = True

        brands = [self.exemplars_brands[idx] for idx in detections["labels"]]
        detections["brands"] = np.array(brands)
        return self._select_detections(detections, selections)

    @staticmethod
    def _select_detections(detections, selections):
        """Specify detections for the given selections.

        :param detections:
        :param selections:
        :return:
        """
        detections["boxes"] = detections["boxes"][selections]
        detections["labels"] = detections["labels"][selections]
        detections["scores"] = detections["scores"][selections]
        detections["brands"] = detections["brands"][selections]
        return detections



from setuptools import setup
from setuptools import find_packages

setup(
    name="logodetect",
    version="1.0.0",
    description="One-shot logo detection for videos and images.",
    long_description="One-shot logo detection for videos and images.",
    url="https://github.com/Heldenkombinat/logodetect",
    download_url="https://github.com/Heldenkombinat/logodetect/tarball/0.1",
    author="Jorge Davila Chacon",
    author_email="jorge@heldenkombinat.com",
    install_requires=[
        "Cython>=0.29.15",
        "click>=7.1.1",
        "imgaug>=0.4.0",
        "matplotlib>=3.2.0rc3",
        "moviepy>=1.0.1",
        "numpy>=1.18.1",
        "opencv-python>=4.2.0.32",
        "pandas>=1.0.1",
        "scipy>=1.4.1",
        "sklearn",
        "torch>=1.4.0",
        "torchvision>=0.5.0",
        "tqdm>=4.42.1",
    ],
    extras_require={
        "tests": ["pytest", "pytest-pep8", "pytest-cov", "mock"],
        "dev": ["black", "pre-commit", "flask", "gunicorn"],
    },
    entry_points={"console_scripts": ["logodetect=logodetect.cli:cli"]},
    packages=find_packages(),
    license="GNU AGPLv3",
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)
